"""
web_server.py

spins up a flask web server to host an api
"""


from flask import Flask, send_file, request
import zipfile 
import io
from .gen_keybinding_img import make_imgs 
from .gen_keybinding_img import this_dir as img_dir
from .discord_log import WALLPAPER_PATH
# from threading import Thread
import shutil
from multiprocessing import Process, Queue
from os.path import expanduser, islink
import os
from libqtile import hook
from libqtile.log_utils import logger
from .auto_desk_api import set_layout
from os import remove as rm, symlink
from .tmux import tmux_layout


app = Flask("frankentile")
APP_HANDLE = None


@app.route("/key-binds", methods=["GET"])
def key_binds():
    """sends back a key bindings image"""
    try:
        shutil.rmtree(img_dir)
    except FileNotFoundError:
        pass

    images = make_imgs(config_path=expanduser("~/.config/qtile/config.py"))

    zip_f = os.path.join(img_dir, "keybinds.zip")

    with zipfile.ZipFile(zip_f, "w", zipfile.ZIP_DEFLATED, False) as zip_file:
        for fname in images:
            zip_file.writestr(fname.split("/")[-1], open(fname, "rb").read())

    return send_file(zip_f, download_name="keybinds.zip", as_attachment=True)


@app.route("/wallpaper", methods=["POST"])
def set_wallpaper():
    """takes a path to an image and sets the wallpaper to it"""
    path = request.values.get("wallpaper-path")

    if islink(WALLPAPER_PATH):
        rm(WALLPAPER_PATH)
        symlink(expanduser(path), WALLPAPER_PATH)

    # TODO: reload configs

    return "success"


@app.route("/auto-desk/<layout>")
def load_layout(layout):
    """uses auto-desk to load the specified layout"""
    res = set_layout(layout)

    return f"setting auto-desk layout {layout}. auto-desk says: {res}"


@app.route("/tmux/<layout>")
async def tmux(layout):
    """sets up a tmux layout"""
    res = await tmux_layout(layout)

    return f"setting up auto-tmux layout {layout}. auto-tmux says: {res}"


def _start_api(host, port):
    logger.warning("running web server")
    app.run(host=host, port=port)  # , debug=True, passthrough_errors=True, use_debugger=False, use_reloader=True)


def start_api(host="127.0.0.1", port=8080):
    """starts the flask server"""
    p = Process(target=_start_api, args=[host, port])
    p.start()
    global APP_HANDLE
    APP_HANDLE = p


def stop_api():
    if APP_HANDLE:
        APP_HANDLE.terminate()


def init():
    hook.subscribe.startup_once(start_api)
    hook.subscribe.shutdown(stop_api)


if __name__ == "__main__":
    start_api()



import discord
import time
import json
import asyncio
import aiohttp
import requests
from discord.ext import commands
from discord import Client
from libqtile.log_utils import logger
from libqtile import hook
from enum import Enum
from os import remove as rm, symlink
from os.path import expanduser, islink
import tomllib
from multiprocessing import Process, Queue
import shutil
import os
from .gen_keybinding_img import make_imgs 
from .gen_keybinding_img import this_dir as img_dir
from .auto_desk_api import set_layout
from .tmux import tmux_layout


COMMAND_PREFIX = "/"
CONFIG_FILE = expanduser("~/.config/qtile/discord.toml")
WALLPAPER_PATH = expanduser("~/.config/qtile/wallpaper")
_intents = discord.Intents.default()
_intents.message_content = True
CLIENT = commands.Bot(intents=_intents, command_prefix=COMMAND_PREFIX)
print(f"global: {CLIENT.status = }")
HANDLES = None


class QueueIter:
    """an iterable wrapper arrounf multiprocessing.Queue"""
    def __init__(self, queue: Queue):
        self.queue = queue  # queue if queue else []

    def append(self, element):
        """add an element to the queue"""
        self.queue.put(element)

    def __repr__(self):
        return self.__str__()
        
    def __str__(self):
        return str(self.queue)

    def __bool__(self):
        return not self.queue.empty()

    def __len__(self):
        return len(self.queue)

    def __iter__(self):
        return self

    def __next__(self):
        if self:
            return self.queue.get()
        else:
            raise StopIteration


class Sender:
    """a state machine that stores messages to be sent to the discord server and sends them on regular intervals"""
    def __init__(self):
        self.connected = False
        self.queue = None
    
    def is_connected(self):
        """returns true if the bot is connected"""
        return self.connected

    async def _send_one(self, session, headers, payload, url):
        """helper function that sends a single message. called by send_all."""
        async with session.post(url, headers=headers, data=payload) as res:
            text = await res.text()

            if res.status != 200:
                logger.warning(f"status: {res.status}, text: {text}")
            
    async def send_all(self):
        """sends all messages in the queue"""
        headers = {
            "Authorization": f"Bot {token}",
            "User-Agent": f"FrankenTile (https://discord.com/developers/applications/1134144480979726446/information v0.1)",
            "Content-Type": "application/json"
        }

        async with aiohttp.ClientSession() as client:
            reqs = []
            
            for id, message in self.queue:
                url = f"https://discord.com/api/v10/channels/{id}/messages"
                payload = json.dumps( {"content": message} )
                reqs.append(self._send_one(client, headers, payload, url))

            await asyncio.gather(*reqs)

    async def try_send_all(self, client, message: (int, str) = None):
        """trys to send messages, if not connected the messages get dumped in the queue"""
        if message:
            self.add_message(message[0], message[1])

        if self.connected:
            await self._send_all(client)

    async def init_sender(self, queue):
        self.queue = QueueIter(queue)

        while True:
            if self.queue:
                await self.send_all()

            await asyncio.sleep(2.5)


MESSENGER = Sender()
QUEUE = Queue()


def init_sender(queue):
    """waits for the bot to be connected and then starts sending messages in an iterval"""
    asyncio.run(MESSENGER.init_sender(queue))


def read_conf():
    """reads the config containing tokens and other information"""
    with open(CONFIG_FILE, "rb") as f:
        return tomllib.load(f)


config = read_conf()
token = config.get("discord").get("token")


class EventType(str, Enum):
    group_switch = "group_switch"
    focus_change = "focus_change"
    window_closed = "window_closed"
    new_window = "new_window"
    restart = "restart"
    resume = "resume"
    mon_change = "mon_change"
    shutdown = "shutdown"
    initial_start = "initial_start"
    booted = "booted"
    login = "login"
    volume = "volume"
    bat_low = "battery_low"
    bluetooth_status = "bluetooth_status"
    keyboard_attached = "keyboard_attached"
    mouse_attached = "mouse_attached"
    keyboard_detached = "keyboard_detached"
    mouse_detached = "mouse_detached"


def bot_start():
    logger.warning("starting discord command bot")
    
    try:
        CLIENT.run(token)
    except Exception as e:
        logger.warning(f"failed to start discord command bot. got error: {e}")
        print(e)


async def send_log(id: int, event_type: EventType, payload: dict):
    """sends a time stamped log to the channel described by id"""
    try:
        text_msg = json.dumps(
            {"timestamp": time.time(), "event_type": event_type, "payload": payload if payload else None}
        )
    except TypeError as e:
        text_msg = f"JSON encoding error: {e}"
        logger.error(f"frankentile error in frankentil.discord.send_log(). {text_msg}")

    await send_mesg(id, text_msg)


async def send_mesg(id: int, mesg: str):
    """sends a timestamped message to the channel described by id"""
    # logger.warning(f"adding to queue {(id, mesg)}")
    QUEUE.put((id, mesg))


async def log(event_type: EventType, payload: dict):
    """generic log function bc many hooks have similar code"""
    try:
        id = config.get("discord").get("log-channel")
    except TypeError as e:
        logger.error(f"could not find log cahnnel id in config file. expected at: \"discord.log-channel\". got error: {e}")
    else:
        await send_log(id, event_type, payload)


@CLIENT.event
async def on_ready():
    msg = f'discord bot {CLIENT.user} is now running!'
    logger.info(msg)
    # print(msg)
    # await log(EventType.login, {"message": "Ready to log!"})


@CLIENT.command(name="get-keybinds")
async def get_keybinds(ctx):
    """replies with the keyboard currently configured""" 
    try:
        shutil.rmtree(img_dir)
    except FileNotFoundError:
        pass

    images = [discord.File(img) for img in make_imgs(config_path=expanduser("~/.config/qtile/config.py"))]
    await ctx.send(files=images)


# @CLIENT.command(name="set-layout")
# async def set_layout(ctx, layout):
#     """uses auto-desk to set the layout"""
#     await ctx.send(f"attempting to set layout {layout} using auto-desk...")
#     # TODO: send layout-load cmd over Unix socket to auto-desk
#     res = set_layout(layout)
#     await ctx.send(f"setup layout {layout}. auto-desk says {res}")


@CLIENT.command(name="set-wallpaper")
async def set_wallpaper(ctx, path):
    """sets the Qtile wallpaper in a semi-permanent fassion"""
    # TODO: add ability to send an image via discord and have this function download it, 
    # save it to ~/.config/qtile/discord_wallpapers/image_name, save the image_name in a log,
    # and finally sent wallpaper to it.  

    await ctx.send(f"setting walpaper to {path}")
    
    if islink(WALLPAPER_PATH):
        rm(WALLPAPER_PATH)
        symlink(expanduser(path), WALLPAPER_PATH)
        ctx.send("set walpaper successfully.")


@CLIENT.command(name="auto-desk")
async def set_wallpaper(ctx, layout):
    """uses auto tmux to spin up a, auto-desk session"""
    await ctx.send(f"loading auto-desk layout {layout}")
    res = set_layout(layout)
    await ctx.send(f"auto-desk said: `{res}`")


@CLIENT.command(name="auto-tmux")
async def set_wallpaper(ctx, layout):
    """uses auto tmux to spin up a tmux session"""
    await ctx.send(f"loading tmux layout {layout}")
    res = await tmux_layout(layout)
    await ctx.send(f"auto-tmux said: `{res}`")


async def closed_window(win):
    """
    Called after a client has been unmanaged
    (when the window is closed by the user.
    """
    await log(
        EventType.window_closed,
        {"name": win.name, "wm_class": win.get_wm_class(), "pid": win.get_pid()}
    )


async def new_window(win):
    """Called after Qtile starts managing a new client."""
    await log(
        EventType.new_window,
        {"name": win.name, "wm_class": win.get_wm_class(), "pid": win.get_pid()}
    )


async def restart():
    """
    Called before Qtile is restarted.
    (hopefully this includes when the device goes to sleep)
    """
    await log(EventType.restart, {})


async def resume():
    """Called when system wakes up from sleep, suspend or hibernate."""
    await log(EventType.resume, {})


async def monitor_change():
    """Called when the output configuration is changed (e.g. via randr in X11)."""
    await log(EventType.mon_change, {})


async def shutdown():
    """Called before Qtile is shutdown"""
    # logger.warning("shutting down")
    await log(EventType.shutdown, {})
    # await kill_bot()


async def start_success():
    """Called when Qtile is started after all resources initialized"""
    await log(EventType.booted, {"message": "Qtile configured successfully"})


async def login():
    """Called when Qtile has started on first start"""
    await log(EventType.login, {"message": "Welcome to Qtile!"})


async def kill_bot():
    """ensures that the bots stop running when the qtile does."""
    logger.info("stopping client bot") 
    await CLIENT.close()


def run_discord_bot():
    """runs the discord bot in parallel""" 
    handles = [
        Process(target=init_sender, args=[QUEUE]),
        Process(target=bot_start, args=()),
    ]

    [h.start() for h in handles]
    
    return handles


def start_discord_bot():
    """starts the discord bot on login"""
    global HANDLES
    HANDLES = run_discord_bot()


async def stop_discord_bot():
    """stops the discord bot on logout."""
    await kill_bot()
    global HANDLES
    [h.terminate() for h in HANDLES]


async def init_logger():
    await MESSENGER.init_sender(QUEUE)
    

def init():
    """initializes the discord api. should be called from Qtile's main config.py"""
    # activate/deactivate bot
    if config.get("discord").get("api"):
        hook.subscribe.startup_once(start_discord_bot)
        hook.subscribe.shutdown(stop_discord_bot)
    else:
        hook.subscribe.startup_once(init_logger)

    # logging functionality
    hook.subscribe.client_killed(closed_window)
    hook.subscribe.client_managed(new_window)
    hook.subscribe.restart(restart)
    hook.subscribe.resume(resume)
    hook.subscribe.screen_change(monitor_change)
    hook.subscribe.shutdown(shutdown)
    hook.subscribe.startup_complete(start_success)
    hook.subscribe.startup_once(login)


if __name__ == "__main__":
    bot_start()



"""
this code is a modified form of the script that can be found at the link below:
https://github.com/qtile/qtile/blob/master/scripts/gen-keybinding-img
"""

#######################################
#  Qtile keybindings image generator  #
#######################################
import os
import sys
import cairocffi as cairo
from cairocffi import ImageSurface


# this_dir = os.path.dirname(__file__)
# base_dir = os.path.abspath(os.path.join(this_dir, ".."))
# sys.path.insert(0, base_dir)

this_dir = "/tmp/qtile-keybindings/"


BUTTON_NAME_Y = 65
BUTTON_NAME_X = 10

COMMAND_Y = 20
COMMAND_X = 10

LEGEND = ['modifiers', 'layout', 'group', 'window', 'other']

CUSTOM_KEYS = {
    'Backspace': 2,
    'Tab': 1.5,
    '\\': 1.5,
    'Return': 2.45,
    'shift': 2,
    'space': 5,
    }


class Button:
    def __init__(self, key, x, y, width, height):
        self.key = key
        self.x = x
        self.y = y
        self.width = width
        self.height = height


class Pos:
    WIDTH = 78
    HEIGHT = 70
    GAP = 5

    def __init__(self, x, y):
        self.x = x
        self.row_x = x
        self.y = y
        self.custom_width = {}
        for i, val in CUSTOM_KEYS.items():
            self.custom_width[i] = val * self.WIDTH

    def get_pos(self, name):
        if name in self.custom_width:
            width = self.custom_width[name]
        else:
            width = self.WIDTH

        info = Button(name, self.x, self.y, width, self.HEIGHT)

        self.x = self.x + self.GAP + width

        return info

    def skip_x(self, times=1):
        self.x = self.x + self.GAP + times * self.WIDTH

    def next_row(self):
        self.x = self.row_x
        self.y = self.y + self.GAP + self.HEIGHT


class KeyboardPNGFactory:
    def __init__(self, modifiers, keys):
        self.keys = keys
        self.modifiers = modifiers.split('-')
        self.key_pos = self.calculate_pos(20, 140)

    def rgb_red(self, context):
        context.set_source_rgb(0.8431372549, 0.3725490196, 0.3725490196)

    def rgb_green(self, context):
        context.set_source_rgb(0.6862745098, 0.6862745098, 0)

    def rgb_yellow(self, context):
        context.set_source_rgb(1, 0.6862745098, 0)

    def rgb_cyan(self, context):
        context.set_source_rgb(0.5137254902, 0.6784313725, 0.6784313725)

    def rgb_violet(self, context):
        context.set_source_rgb(0.831372549, 0.5215686275, 0.6784313725)

    def calculate_pos(self, x, y):
        pos = Pos(x, y)

        key_pos = {}
        for c in '`1234567890-=':
            key_pos[c] = pos.get_pos(c)

        key_pos['Backspace'] = pos.get_pos('Backspace')
        pos.next_row()

        key_pos['Tab'] = pos.get_pos('Tab')
        for c in 'qwertyuiop[]\\':
            key_pos[c] = pos.get_pos(c)
        pos.next_row()

        pos.skip_x(1.6)
        for c in 'asdfghjkl;\'':
            key_pos[c] = pos.get_pos(c)
        key_pos['Return'] = pos.get_pos('Return')
        pos.next_row()

        key_pos['shift'] = pos.get_pos('shift')
        for c in 'zxcvbnm':
            key_pos[c] = pos.get_pos(c)
        key_pos['period'] = pos.get_pos('period')
        key_pos['comma'] = pos.get_pos('comma')
        key_pos['/'] = pos.get_pos('/')
        pos.next_row()

        key_pos['control'] = pos.get_pos('control')
        pos.skip_x()
        key_pos['mod4'] = pos.get_pos('mod4')
        key_pos['mod1'] = pos.get_pos('mod1')
        key_pos['space'] = pos.get_pos('space')
        key_pos['Print'] = pos.get_pos('Print')
        pos.skip_x(3)
        key_pos['Up'] = pos.get_pos('Up')

        pos.next_row()
        pos.skip_x(12.33)
        key_pos['Left'] = pos.get_pos('Left')
        key_pos['Down'] = pos.get_pos('Down')
        key_pos['Right'] = pos.get_pos('Right')

        pos.next_row()

        for l in LEGEND:
            key_pos[l] = pos.get_pos(l)

        pos.skip_x(5)
        key_pos['Button1'] = pos.get_pos('Button1')
        key_pos['Button2'] = pos.get_pos('Button2')
        key_pos['Button3'] = pos.get_pos('Button3')

        pos.next_row()
        key_pos['FN_KEYS'] = pos.get_pos('FN_KEYS')

        return key_pos

    def add_logo(self, context):
        logo_img = os.path.abspath(os.path.join(this_dir, "..", 'logo.png'))

        context.save()
        context.scale(0.5)
        logo = ImageSurface.create_from_png(logo_img)
        context.set_source_surface(logo, 20, 50)
        context.paint()
        context.restore()

    def render(self, filename):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 1280, 800)
        context = cairo.Context(surface)
        with context:
            context.set_source_rgb(1, 1, 1)
            context.paint()

        # self.add_logo(context)

        context.move_to(210, 80)
        context.set_font_size(28)
        context.show_text('Keybindings for Qtile')

        context.move_to(210, 100)
        context.set_font_size(18)
        if len([i for i in self.modifiers if i]):
            context.show_text('Modifiers: ' + ', '.join(self.modifiers))
        else:
            context.show_text('No modifiers used.')

        for i in self.key_pos.values():
            if i.key in ['FN_KEYS']:
                continue

            self.draw_button(context, i.key, i.x, i.y, i.width, i.height)

        # draw functional
        fn = [i for i in self.keys.values() if i.key[:4] == 'XF86']
        if len(fn):
            fn_pos = self.key_pos['FN_KEYS']
            x = fn_pos.x
            for i in fn:
                self.draw_button(context, i.key, x, fn_pos.y, fn_pos.width, fn_pos.height)
                x += Pos.GAP + Pos.WIDTH

        # draw mouse base
        context.rectangle(830, 660, 244, 90)
        context.set_source_rgb(0, 0, 0)
        context.stroke()
        context.set_font_size(28)
        context.move_to(900, 720)
        context.show_text('MOUSE')

        surface.write_to_png(filename)

    def draw_button(self, context, key, x, y, width, height):
        fn = False
        if key[:4] == 'XF86':
            fn = True

        if key in LEGEND:
            if key == 'modifiers':
                self.rgb_red(context)
            elif key == 'group':
                self.rgb_green(context)
            elif key == 'layout':
                self.rgb_cyan(context)
            elif key == 'window':
                self.rgb_yellow(context)
            else:
                self.rgb_violet(context)
            context.rectangle(x, y, width, height)
            context.fill()

        if key in self.modifiers:
            context.rectangle(x, y, width, height)
            self.rgb_red(context)
            context.fill()

        if key in self.keys:
            k = self.keys[key]
            context.rectangle(x, y, width, height)
            self.set_key_color(context, k)
            context.fill()

            self.show_multiline(context, x + COMMAND_X, y + COMMAND_Y, k)

        context.rectangle(x, y, width, height)
        context.set_source_rgb(0, 0, 0)
        context.stroke()

        if fn:
            key = key[4:]
            context.set_font_size(10)
        else:
            context.set_font_size(14)

        context.move_to(x + BUTTON_NAME_X, y + BUTTON_NAME_Y)
        context.show_text(self.translate(key))

    def show_multiline(self, context, x, y, key):
        """Cairo doesn't support multiline. Added with word wrapping.
        """
        c_width = 14
        if key.key in CUSTOM_KEYS:
            c_width *= CUSTOM_KEYS[key.key]

        context.set_font_size(10)
        context.set_source_rgb(0, 0, 0)
        context.move_to(x, y)
        words = key.command.split(' ')
        words.reverse()
        printable = last_word = words.pop()
        while len(words):
            last_word = words.pop()
            if len(printable + ' ' + last_word) < c_width:
                printable += ' ' + last_word
                continue

            context.show_text(printable)
            y += 10
            context.move_to(x, y)
            printable = last_word

        if last_word is not None:
            context.show_text(printable)

    def set_key_color(self, context, key):
        if key.scope == 'group':
            self.rgb_green(context)
        elif key.scope == 'layout':
            self.rgb_cyan(context)
        elif key.scope == 'window':
            self.rgb_yellow(context)
        else:
            self.rgb_violet(context)

    def translate(self, text):
        dictionary = {
            'period': ',',
            'comma': '.',
            'Left': '←',
            'Down': '↓',
            'Right': '→',
            'Up': '↑',
            'AudioRaiseVolume': 'Volume up',
            'AudioLowerVolume': 'Volume down',
            'AudioMute': 'Audio mute',
            'AudioMicMute': 'Mic mute',
            'MonBrightnessUp': 'Brightness up',
            'MonBrightnessDown': 'Brightness down',
            }

        if text not in dictionary:
            return text

        return dictionary[text]


class KInfo:
    NAME_MAP = {
            'togroup': 'to group',
            'toscreen': 'to screen',
            }

    def __init__(self, key):
        self.key = key.key
        self.command = self.get_command(key)
        self.scope = self.get_scope(key)

    def get_command(self, key):
        if hasattr(key, 'desc') and key.desc:
            return key.desc

        cmd = key.commands[0]
        command = cmd.name
        if command in self.NAME_MAP:
            command = self.NAME_MAP[command]

        command = command.replace('_', ' ')

        if len(cmd.args):
            if isinstance(cmd.args[0], str):
                command += ' ' + cmd.args[0]

        return command

    def get_scope(self, key):
        selectors = key.commands[0].selectors
        if len(selectors):
            return selectors[0][0]


class MInfo(KInfo):
    def __init__(self, mouse):
        self.key = mouse.button
        self.command = self.get_command(mouse)
        self.scope = self.get_scope(mouse)


def get_kb_map(config_path):
    from libqtile.confreader import Config

    c = Config(config_path)
    if config_path:
        c.load() 

    kb_map = {}
    for key in c.keys:
        mod = '-'.join(key.modifiers)
        if mod not in kb_map:
            kb_map[mod] = {}

        info = KInfo(key)
        kb_map[mod][info.key] = info

    for mouse in c.mouse:
        mod = '-'.join(mouse.modifiers)
        if mod not in kb_map:
            kb_map[mod] = {}

        info = MInfo(mouse)
        kb_map[mod][info.key] = info

    return kb_map



def make_imgs(config_path=None):
    # config_path = None
    output_dir = this_dir
    os.makedirs(output_dir, exist_ok=True)
    files = []

    kb_map = get_kb_map(config_path if config_path else os.expanduser("~/.config/qtile/config.py"))
    for modifier, keys in kb_map.items():
        if not modifier:
            filename = "no_modifier.png"
        else:
            filename = "{}.png".format(modifier)

        output_file = os.path.abspath(os.path.join(output_dir, filename))
        f = KeyboardPNGFactory(modifier, keys)
        f.render(output_file)
        files.append(output_file)

    return files


if __name__ == '__main__':
    make_imgs()



#####################################
# Auto-Desk API stuff, helper hooks #
#####################################



from libqtile.command.client import InteractiveCommandClient
from libqtile.log_utils import logger
from libqtile import hook
from socket import socket, AF_UNIX, SOCK_STREAM


QTILE_CLIENT = InteractiveCommandClient()
NEW_CLIENT_PIDs = set()
PATH = "/tmp/desktop-automater"


def _open_on(client):
    """used to move windows when they open""" 
    pid = client.get_pid()

    # this function gets called twice per window opening.
    # bellow stops this function from moving windows that have already been moved.
    if pid not in NEW_CLIENT_PIDs:
        NEW_CLIENT_PIDs.add(pid)
        move_window(client)
    else:
        NEW_CLIENT_PIDs.remove(pid)


# @hook.subscribe.client_managed
async def open_on_backup(client):
    """
    used to move windows when the program sets its WM_CLASS after its managed (ie, after its registered)
    this was made bc spotify doesn't like playing nice with linux.

    this a back up for open_on().
    """
    # logger.warning("new client (from backup function)")
    _open_on(client)


# @hook.subscribe.client_new
async def open_on(client):
    """moves windows when they register"""
    # logger.warning("new client")
    _open_on(client)


# @hook.subscribe.group_window_add
async def clear_group(group, window):
    # logger.warning(f"clearing group \"{group.name}\"")
    clearing = should_clear(group.name)
    if clearing:
        logger.debug(f"clearing group {group.name}")
        pid = window.get_pid()
        for w in group.windows:
            if w.get_pid() != pid:
                w.togroup("hidden")


def get_location(wm_class):
    message = f"auto-move {wm_class[0]} {wm_class[1]}"
    return send_auto_desk(message)


def should_clear(group):
    message = f"should-clear {group}"
    res = send_auto_desk(message)
    logger.debug(f"should-clear res: '{res}'")
    return res == "true"


def send_auto_desk(message):
    """sends data to auto-desk and returns the response"""
    location = None

    with socket(AF_UNIX, SOCK_STREAM) as s:
        s.settimeout(10)
        try:
            s.connect(PATH)
        except FileNotFoundError:
            pass
        except TimeoutError:
            pass
        else:
            s.send(bytes(message, "utf-8"))
            s.shutdown(1)  # tells the server im done sending data and it can reply now.
            res = s.recv(1024)
            ec = res[0]
            if len(res) >= 3:
                location = res[2:].decode('utf-8')
            if ec:
                logger.error(f"got error code from auto-desk on message '{message}'.")

    return location


# untested
def clear_desktop(group):
    """clears all desktop in self.clears"""
    # for group in tmp_clears:
    if group:
        # windows = [w["id"]
                # for w in QTILE_CLIENT.windows() if w["group"] == group]
        windows = QTILE_CLIENT.group[group].windows()
        logger.info(f"about to clear windows from group '{group}'")
        for wid in windows:
            QTILE_CLIENT.window[wid].togroup("hidden")
        logger.info(f"cleared group '{group}'")
    else:
        logger.info(f"not clearing group '{group}'")            


def move_window(c):
    wm_class = c.get_wm_class()
    location = get_location(wm_class)
    logger.debug(f"moving to location, '{location}'")
    # clear = should_clear(location)
    # if clear:
    #     clear_desktop(location)
    if location:
        c.togroup(location)


def init():
    """intializes the auto-desk api. should be called once from Qtile's config.py"""
    hook.subscribe.client_new(open_on)
    hook.subscribe.client_managed(open_on_backup)
    hook.subscribe.group_window_add(clear_group)



import libtmux
from auto_tmux import _get_full_path, load_layout
from tqdm.contrib.logging import logging_redirect_tqdm
from os.path import isfile
from libqtile.log_utils import logger


async def tmux_layout(name):
    """loads the layout"""
    logger.warning(f"name :  {name}")
    layout_path = _get_full_path(name)

    logger.warning(f"layout :  {layout_path}")
    if not isfile(layout_path):
        return "not a valid layout"
    
    server = libtmux.Server()
    layout = await load_layout(server, layout_path, progress_bar=False)

    return "loaded succefully"



"""
a simple api for interacting with auto-desk
"""


from os.path import expanduser
import socket
import tomllib


HOME = expanduser("~")
_CONF_FILE = f"{HOME}/.config/auto-desk/config.toml"


def configs():
    with open(_CONF_FILE, "rb") as f:
        data = tomllib.load(f)
        return data


def set_layout(layout):
    path = configs().get("server").get("listen_socket")

    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as s:
        s.connect(path)
        s.send(f"load-layout {layout}".encode("ascii"))

        s.shutdown(1)
        res = s.recv(1024).decode('utf-8')

        return res

